#
# Automatic component or layout simplification tool for animation exporting
# 2021.03.11
#

from vcCommand import *
import math

app = getApplication()
cmd = getCommand()
max_components = 0
cur_components = 0
  
target_percentage = 25.0
geometry_minimum = 0.5
red = app.findMaterial("red")
yellow = app.findMaterial("yellow")

MAX_ITERATIONS = 10

def processTriangleSet(set):
    
    origTriCount = set.TriangleCount
    prevTriCount = origTriCount
    triCount = origTriCount
    tolerance = 0.1

    # Decimate first with tight tolerance to cleanup highly tesselated surfaces
    # set.decimate(-10.0, 0.001)

    target_percentage = 100.0
    if triCount > 24:
        target_percentage = 75.0
    if triCount > 50:
        target_percentage = 65.0
    if triCount > 100:
        target_percentage = 55.0
    if triCount > 200:
        target_percentage = 50.0
    if triCount > 500:
        target_percentage = 45.0
    if triCount > 2000:
        target_percentage = 40.0
    if triCount > 5000:
        target_percentage = 37.0
    if triCount > 10000:
        target_percentage = 35.0
    if triCount > 25000:
        target_percentage = 30.0 # 30.0
    if triCount > 50000:
        target_percentage = 26.0 # 26.0
    if triCount > 100000:
        target_percentage = 22.0 # 23.0
    if triCount > 250000:
        target_percentage = 20.0 # 20.0
    if triCount > 1000000:
        target_percentage = 15.0 # 15.0


    radius = set.BoundDiagonal.length()
    if radius < geometry_minimum:
        return

    max_tolerance = radius*0.125*(target_percentage/100.0)
    tolerance = 0.001*radius
    
    # Adjust max tolerance for geometrically bigger items larger
    if radius > 0.1:
      scale_size = 1.0 + radius/2000.0
      #max_tolerance = scale_size*5.0*(1.0 - target_percentage/100.0)+0.00001
      max_tolerance = scale_size*1.50*(1.0 - target_percentage/100.0)+0.00001
      
    if intensity_prop.Value == 'Careful 75%':
        target_percentage = min(target_percentage*1.75, 100)
        max_tolerance *= 0.5
    elif intensity_prop.Value == 'Strong 25%':
        target_percentage = target_percentage*0.10
        max_tolerance *= 1.2
    elif intensity_prop.Value == 'Coarse 5%':
        target_percentage = target_percentage*0.05
        max_tolerance *= 5.0


    # Tolerance value is used ins steps to reduce geometry until max tolerance or target reduction is reached
    tolerance = max_tolerance # max_tolerance*0.25

    if target_percentage >= 100.0:
        triCount = set.TriangleCount
        set.update()
        return [origTriCount,triCount]
    last_percentage = 100.0
    iterations = 0
    while iterations < MAX_ITERATIONS:
        # Negative target percentage is prototype option for decimator to use point merging before decimation to improve quality (reduces tearing geometry apart)
        set.decimate(-target_percentage, max_tolerance)
        triCount = set.TriangleCount
        if triCount == prevTriCount:
            # print "  Target cannot be achieved (cur) tolerance:",tolerance,target_percentage," (",triCount,")"
            break
        percentage = 100*float(triCount)/float(origTriCount) 
        if percentage < target_percentage:
            # print "  Target achieved (tolerance:",tolerance,")",target_percentage," (",triCount,")"
            break     
        if (last_percentage-percentage) < 0.1:
            # print "  Target converging too slow - abort (tolerance:",tolerance,")",target_percentage," (",triCount,")"
            break   
        last_percentage = percentage
        if tolerance > max_tolerance:
            # print "  Target achieved with (max) tolerance:",tolerance, target_percentage
            # Debug material
            # set.Material = red
            break
        prevTriCount = triCount
        break
        if abs(percentage-target_percentage) < 10.0:
            tolerance *= 1.25
        else:
            tolerance *= 200.0

        iterations += 1

    # Testing
    #if triCount < 10000:
    #    set.Material = yellow

    set.update()
    return [origTriCount,triCount]


def ProcessNodes(node, comp):
    global comp_sum_origs, comp_sum_tris
    global processed_containers, processed_counts
    if node.Component != comp:
        return

    # Process this nodes geometry
    container = node.Geometry
    name = comp.Name + ":"
    name += node.Name
    try:
        index = processed_containers.index(container)
        sum_origs,sum_tris = processed_counts[index]
        comp_sum_origs += sum_origs
        comp_sum_tris += sum_tris
        print "  Skipping already processed container for component:",name
        return
    except:
        pass

    processed_containers.append(container)
    sum_origs = 0
    sum_tris = 0
    for set in container.GeometrySets:
        try:
            if set.Type == VC_TRIANGLESET:
                orig_tris, tris = processTriangleSet(set)
                sum_origs += orig_tris
                sum_tris += tris
        except:
            pass

    comp_sum_origs += sum_origs
    comp_sum_tris += sum_tris

    processed_counts.append([sum_origs,sum_tris])

    # Process all child nodes if they belong to this component
    for child in node.Children:
        ProcessNodes(child, comp)


def ProcessComponentFeeders(comp):
    feeders = comp.getBehavioursByType(VC_COMPONENTCREATOR)
    for feeder in feeders:
        if feeder.TemplateComponent:
            ProcessComponent(feeder.TemplateComponent)

def ProcessComponent(comp):
    ProcessNodes(comp, comp)
    ProcessComponentFeeders(comp)
    app.render()

def get_scope_comps():
    comps = []
    if include_prop.Value == 'Selected Components':
        selection = app.SelectionManager
        if len(selection.getSelection(VC_SELECTION_COMPONENT)) == 0:
            print "WARNING: Select at least one component to simplify"
            return comps
        comps = selection.getSelection(VC_SELECTION_COMPONENT)
    else:
        comps = app.Components
    return comps

def process(arg):
    global comp_sum_origs, comp_sum_tris
    global processed_containers, processed_counts
    global max_components, cur_components
    comps = get_scope_comps()
    if not comps or len(comps) == 0:
      return

    max_components = len(comps)
    cur_components = 0
    app.ProgressStatus = "Simplifying"
    app.ProgressValue = float(cur_components)/float(max_components)

    processed_containers = []
    processed_counts = []
    total_comp_sum_origs = 0
    total_comp_sum_tris = 0
    for comp in comps:
        title = "Simplifying '"+comp.Name+"'"
        app.ProgressStatus = title
        app.ProgressValue = float(cur_components)/float(max_components)
        print "SIMPLIFYING:",comp.Name
        comp_sum_origs = 0
        comp_sum_tris = 0
        ProcessComponent(comp)
        cur_components += 1
        if comp_sum_origs > 0:
            reduction = (100*float(comp_sum_tris)/float(comp_sum_origs))
            print "  Simplification percentage: ",reduction, " Triangles(Original:",comp_sum_origs,"Simplified:",comp_sum_tris,")"
            total_comp_sum_origs += comp_sum_origs
            total_comp_sum_tris += comp_sum_tris
        comp.rebuild()#202403 Tang add
    app.ProgressValue = 2.0
    if total_comp_sum_origs > 0:
        reduction = (100*float(total_comp_sum_tris)/float(total_comp_sum_origs))
        print "TOTAL simplification percentage: ",reduction, " Triangles(Original:",total_comp_sum_origs,"Simplified:",total_comp_sum_tris,")"


def reset(arg):
    comps = get_scope_comps()
    max_components = len(comps)
    cur_components = 0
    app.ProgressStatus = "Reset simplification"
    app.ProgressValue = float(cur_components)/float(max_components)

    for comp in comps:
        if comp.Material:
            cached_material = comp.Material
            comp.Material = None
            comp.Material = cached_material
        else:
            if not red:
                print 'warning: material lib issue. red not found'
            comp.Material = red
            comp.Material = None
        cur_components += 1
        app.ProgressValue = float(cur_components)/float(max_components)
    app.ProgressValue = 2.0
    app.render()


def OnStart():
    global include_prop, intensity_prop, simplify_prop, reset_prop
    include_prop = createProperty(VC_STRING,'Include', VC_PROPERTY_STEP)
    include_prop.StepValues = ['Selected Components', 'Whole Layout']
    include_prop.Value = include_prop.StepValues[0]

    intensity_prop = createProperty(VC_STRING,'Target', VC_PROPERTY_STEP)
    intensity_prop.StepValues = ['Careful 75%', 'Medium 50%', 'Strong 25%', 'Coarse 5%']
    intensity_prop.Value = intensity_prop.StepValues[1]

    simplify_prop = createProperty(VC_BUTTON,'Simplify')
    simplify_prop.OnChanged = process

    reset_prop = createProperty(VC_BUTTON,'Reset')
    reset_prop.OnChanged = reset

    executeInActionPanel()



addState( None )
